/******************************************************************************

  Copyright (c) 1999-2010 by Turku PET Centre

  Program:     fit_srtm
  Description: Non-linear fitting of the parameters of simplified reference 
               tissue compartmental model to regional PET time-activity curves.

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation; either version 2 of the License, or (at your option) any later
  version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 Temple
  Place, Suite 330, Boston, MA 02111-1307 USA.

  Turku PET Centre hereby disclaims all copyright interest in the program.

  Juhani Knuuti
  Director, Professor
  Turku PET Centre, Turku, Finland, http://www.turkupetcentre.fi

  Modification history:
  1.0   1999-04-29 Vesa Oikonen
        First created. Based on programs fithume and fitbp (by VO).
  1.1   2000-01-10 VO
        y2k correction.
  1.2   2000-04-11 JR
        Reading and Writing procedures changed to DFT format
        Possibility to use separate reference TAC file is added
  1.3   2000-05-12 VO
        Name of fit_rtcmr changed to fit_rrtm (reduced ref tissue model).
  1.4   2000-07-19 JR
        interp.c replaced with integr.c
        Program uses libpet library.
  1.5   2000-08-01 CL
        stdout is now flushed after every printf
        (this is required by the GUI)
  1.6   2000-08-15 VO
        Included 4th optional program fit_trtm.
        Defining UNIX/DOS selection removed.
  1.7   2000-10-05 VO
        If datafile contains the weights, they are applied in the fitting.
        Corrected a possible memory overflow with p[][].
        2000-12-13 VO
        Minor changes not affecting compiled code.
  1.8   2001-06-12 VO
        pivot() used instead of pivot_lep(); other changes linked into that.
        2001-06-13 VO
        Corrected a bug in selecting ref region.
        2001-07-03 VO
        Simulation functions in libpet applied.
        For that reason, integrations and calculations of dt2 are removed,
        but time pointer is added.
        2001-07-04 VO
        Datasets for integrated data are removed from the main function.
        Warning from high parameter sd is now given only if sd>20%.
        More test output.
        2001-07-09 VO
        trtmSIM was changed in petlib; required changes here.
  1.9   2001-08-08 VO
        The probeNr in pivot() is changed by setting PIVOT_PROBES_PER_PAR
        to a higher value in here.
        Initial BP for fit is calculated from the individual roi-to-ref ratio.
        Max BP is changed from 50 to 30 for rtcm and srtm models.
        2001-08-15 VO
        Max BP is set individually as 10x roi-to-ref ratio, with 30 as the
        absolute maximum.
        2001-09-11 VO
        Min and max BP is set individually as 0.2-4x roi-to-ref ratio minus 1,
        with 30 as the absolute maximum.
  1.10  2001-09-12 VO
        Included option -i<file> for specifying initial values and constraints.
        Necessary changes in argument handling.
  2.0   2002-11-14 VO
        pivot() replaced by tgo(), as required by change in libpet.
        Result data structure and functions in libpet are applied.
        Corrected a bug in setting automatic BP limits when (AUC ratio - 1) was
        negative.
        Changes in info texts.
  2.1   2003-02-19 VO
        Computes 95% confidence limits or SD's with bootstrap.
        Default upper limits for R1 were changed from 5.0 to 10.0.
  2.2   2003-10-30 VO
        Change in setting automatic BP upper limit from (AUC ratio - 1);
        previously 4x, now 5x that.
        Removed unnecessary include.
  2.3   2004-03-17 VO
        Change in setting automatic BP upper limit:
        Now r=AUC ratio; if r<1.0, then r=1.0; upper limit is set to 5*r.
  2.3.1 2004-08-10 VO
        Changes in processing arguments:
        Now both -CL and -SD can be given.
        Accepts options -H and -V.
        Inifile can be given with '='.
  2.4.0 2005-07-29 VO
        The code for this model is separated from sources (fit_rtcm.c) to all
        reference tissue compartmental models (FRTM, SRTM, RRTM, TRTM).
        Applies new library versions.
        Optimization is done by calling powell() two times, not using tgo().
        Library function is used in calculation of AIC; previosly WSS was not
        divided by nr of data points as it should have been.
        SC is not calculated or reported anymore.
        Lots of changes in comments and info text
        Lots of changes in processing command line options.
        Gives warning if reference ROI file contains more than one TAC.
        Frame with zero time is never added to data.
        No more checking of estimates being close to limits.
  2.4.1 2005-08-01 VO
        Corrected a bug in checking nr of TACs in reference tissue file.
        Program name is written as comment with fitted TACs.
  2.4.2 2007-12-30 VO
        Makefile changed to link the most recent library versions.
        No other changes in this file than includes and this mod history.
        Version number changed because result file format is now changing.
  3.0.0 2008-01-17 VO
        Licence changed from LGPL to GPL.
        User info updated.
        libtpcmodext functions used to set results contents.
        Time units in results corrected.
        Accepts options --version, --build, --history and --verbose.
        Option -fit=filename added; corresponding optional comman-line argument
        removed.
        Option -n removed (meant that neither SD or CL is estimated).
        Options -CL and -SD are changed to -CL[=Y|n] and -SD[=Y|n]; 
        By default, neither is estimated.
        When parameters exceed limits, fit is penalized, not set to BAD_FIT. 
        Added option -svg=filename.
        Added option -DVR
        AIC calculation takes into account if any parameter is fixed.
        Reference curve is read using library function.
        Fit duration is determined using library function.
        Fit constraint file format and i/o changed.
        Fit duration (command-line argument) must be > 0.
        Default upper limit of k2 increased from 1 to 10.
  3.0.1 2008-06-30 VO
        Uses iTGO instead of 2 x Powell-Brent for fitting.
        Small changes in user info and error messages etc.
        Option -i changed to -lim, although -i is still working.
        Changes in how reference is read.
  3.0.2 2009-06-09 VO
        No changes in fit_srtm actually, but new library versions needed
        additional changes to make this program pass the tests.
  3.0.3 2010-04-26 VO
        Studynr is written in SVG title.
        Small changes in user help text.
  3.0.4 2010-09-06 VO
        Penalty is not applied to WSS and AIC.
        Saved result parameters are kept withing constraints.
  3.0.5 2010-09-07 VO
        dftReadReference() used in reading reference input.
  3.0.6 2010-11-10 VO
        NA definition replaced by NaN (isnan(f) and f=nan("")).



 
*******************************************************************************/
#define PROGRAM "fit_srtm 3.0.6  (c) 1999-2010 by Turku PET Centre"
/******************************************************************************/
#define MAX_PARAMS 3
/******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
/******************************************************************************/
#include "curveio.h"
#include "libtpcmodext.h"
#include "libtpcmisc.h"
#include "libtpcmodel.h"
#include "libtpcsvg.h"
/******************************************************************************/
#ifndef BAD_FIT
#define BAD_FIT 9.99e19
#endif
/******************************************************************************/
int TEST=0;
int SILENT=0;
int VERBOSE=0;
/******************************************************************************/
double srtm_func(double *p);
/* Local functions */
void print_usage();
void print_build();
/******************************************************************************/
int fitframeNr=0, parNr=3;
double *t, *cr, *ct, *tis, *w; /* These are pointers, not allocated */
double pinit[MAX_PARAMS], pmin[MAX_PARAMS], pmax[MAX_PARAMS];
double wss_wo_penalty=0.0;
/******************************************************************************/

/******************************************************************************
 *
 *  Main
 *
 ******************************************************************************/
int main(int argc, char **argv)
{
  int           ai, ri, fi, pi, n=0, ret, ref=-1, inputtype;
  int           doSD=0, doCL=0, doBootstrap=0, doDVR=0;
  DFT           dft;
  IFT           ift;
  double        starttime, endtime;
  int           first, last;
  double        a, b, c, refIntegral;
  double       *p, fitdur=-1.0, wss, aic, *sd, *cl1, *cl2;
  char         *cptr, tmp[256],
                dftfile[FILENAME_MAX], svgfile[FILENAME_MAX],
                fitfile[FILENAME_MAX], limfile[FILENAME_MAX],
                reffile[FILENAME_MAX], resfile[FILENAME_MAX];
  RES           res;
  double        def_pmin[MAX_PARAMETERS], def_pmax[MAX_PARAMETERS];

  /* 
   *  Set parameter constraints
   */
  /* R1  */ def_pmin[0]=0.001;     def_pmax[0]=10.0;
  /* k2  */ def_pmin[1]=0.000001;  def_pmax[1]=10.0;
  /* BP  */ def_pmin[2]=0.0;       def_pmax[2]=60.0; // may be reset later

  /*
   *  Get and check arguments
   */
  if(argc==1) {print_usage(); return(1);}
  iftInit(&ift);
  dftfile[0]=fitfile[0]=limfile[0]=reffile[0]=resfile[0]=svgfile[0]=(char)0;
  for(ai=1; ai<argc; ai++) if(*argv[ai]=='-') {
    cptr=argv[ai]+1; if(*cptr=='-') cptr++; if(cptr==NULL) continue;
    if(strncasecmp(cptr, "TEST", 4)==0) {
      cptr+=4; if(*cptr=='=') cptr++; TEST=atoi(cptr); if(TEST<1) TEST=1;
      continue;
    } else if(strcasecmp(cptr, "V")==0 || strcasecmp(cptr, "VERSION")==0 ||
              strcasecmp(cptr, "BUILD")==0) {
      print_build();
      return(0);
    } else if(strcasecmp(cptr, "H")==0 || strcasecmp(cptr, "HELP")==0) {
      print_usage(); return(0);
    } else if(strcasecmp(cptr, "SILENT")==0) {
      SILENT=1; VERBOSE=0; continue;
    } else if(strcasecmp(cptr, "VERBOSE")==0) {
      VERBOSE=1; SILENT=0; continue;
    } else if(strncasecmp(cptr, "LIM=", 4)==0 && strlen(cptr)>4) {
      strcpy(limfile, cptr+4); continue;
    } else if(strncasecmp(cptr, "I=", 2)==0 && strlen(cptr)>2) {
      strcpy(limfile, cptr+2); continue;
    } else if(strncasecmp(cptr, "CL", 2)==0) {
      if(strlen(cptr)==2) {doCL=1; continue;}
      cptr+=2; if(*cptr=='=') {
        cptr++;
        if(*cptr=='Y' || *cptr=='y') {doCL=1; continue;}
        if(*cptr=='N' || *cptr=='n') {doCL=0; continue;}
      }
    } else if(strncasecmp(cptr, "SD", 2)==0) {
      if(strlen(cptr)==2) {doSD=1; continue;}
      cptr+=2; if(*cptr=='=') {
        cptr++;
        if(*cptr=='Y' || *cptr=='y') {doSD=1; continue;}
        if(*cptr=='N' || *cptr=='n') {doSD=0; continue;}
      }
    } else if(strcasecmp(cptr, "DVR")==0) {
      doDVR=1; continue;
    } else if(strncasecmp(cptr, "FIT=", 4)==0) {
      strcpy(fitfile, cptr+4); if(strlen(fitfile)>0) continue;
    } else if(strncasecmp(cptr, "SVG=", 4)==0) {
      strcpy(svgfile, cptr+4); if(strlen(svgfile)>0) continue;
    }
    /* we should never get this far */
    fprintf(stderr, "Error: invalid option '%s'.\n", argv[ai]);
    return(1);
  } else {
    if(!dftfile[0]) {
      strcpy(dftfile, argv[ai]); continue;
    } else if(!reffile[0]) {
      strcpy(reffile, argv[ai]); continue;
    } else if(fitdur<0) {
      fitdur=atof_dpi(argv[ai]);
      if(fitdur<=0.0) {
        fprintf(stderr, "Error: invalid fit time: '%s'.\n", argv[ai]);
        return(1);
      }
      continue;
    } else if(!resfile[0]) {
      strcpy(resfile, argv[ai]); continue;
    }
    fprintf(stderr, "Error: too many arguments: '%s'.\n", argv[ai]);
    return(1);
  }
  if(doSD || doCL) doBootstrap=1; else doBootstrap=0;
  if(TEST) {
    printf("TEST := %d\n", TEST);
    printf("SILENT := %d\n", SILENT);
    printf("VERBOSE := %d\n", VERBOSE);
    printf("dftfile := %s\n", dftfile);
    printf("reffile := %s\n", reffile);
    printf("resfile := %s\n", resfile);
    printf("fitfile := %s\n", fitfile);
    printf("svgfile := %s\n", svgfile);
    printf("required_fitdur := %g\n", fitdur);
    printf("limfile := %s\n", limfile);
    printf("doSD := %d\ndoCL := %d\ndoBootstrap := %d\n", 
      doSD, doCL, doBootstrap);
    printf("doDVR := %d\n", doDVR);
  }
  BS_TEST=0; if(TEST==9) BS_TEST=1;
  if(TEST>8) MODEXT_TEST=TEST-8; else MODEXT_TEST=0;
  if(TEST>7) TGO_TEST=TEST-7; else TGO_TEST=0;
  if(TEST>0) VERBOSE=1;


  /* If only filename for fit limits was given, then write one
     with default contents and exit */
  if(limfile[0] && !dftfile[0]) {
    if(VERBOSE) printf("writing parameter constraints file\n");
    /* Check that constraint file does not exist */
    if(access(limfile, 0) != -1) {
      fprintf(stderr, "Error: constraint file '%s' exists.\n", limfile);
      return(9);
    }
    /* Create parameter file */
    sprintf(tmp, "%g", def_pmin[0]); iftPut(&ift, "R1_lower", tmp, " ");
    sprintf(tmp, "%g", def_pmax[0]); iftPut(&ift, "R1_upper", tmp, " ");
    sprintf(tmp, "%g", def_pmin[1]); iftPut(&ift, "k2_lower", tmp, " ");
    sprintf(tmp, "%g", def_pmax[1]); iftPut(&ift, "k2_upper", tmp, " ");
    sprintf(tmp, "%g", def_pmin[2]); iftPut(&ift, "BP_lower", tmp, " ");
    sprintf(tmp, "%g", def_pmax[2]); iftPut(&ift, "BP_upper", tmp, " ");
    ret=iftWrite(&ift, limfile);
    if(ret) {
      fprintf(stderr, "Error in writing '%s': %s\n", limfile, ift.status);
      return(9);
    }
    if(strcasecmp(limfile, "stdout")!=0)
      if(SILENT==0)
        fprintf(stdout, "File %s with fit limits written.\n", limfile);
    iftEmpty(&ift); return(0);
  }

  /* Did we get all the information that we need? */
  if(fitdur==0) fitdur=99999.; else if(fitdur<0) {print_usage(); return(1);}
  if(!resfile[0]) {
    fprintf(stderr, "Error: missing command-line argument; try %s --help\n",
      argv[0]);
    return(1);
  }


  /*
   *  Read initial values from limfile, if provided
   */
  if(limfile[0]) {
    if(VERBOSE) printf("reading constraints from %s\n", limfile);
    ret=iftRead(&ift, limfile, 1);
    if(ret) {
      fprintf(stderr, "Error in reading '%s': %s\n", limfile, ift.status);
      return(9);
    }
    if(TEST>6) iftWrite(&ift, "stdout");
    n=0;
    /* R1 */
    strcpy(tmp, "R1_lower"); ai=iftGet(&ift, tmp);
    if(ai>=0) {def_pmin[0]=atof_dpi(ift.item[ai].value); n++;}
    strcpy(tmp, "R1_upper"); ai=iftGet(&ift, tmp);
    if(ai>=0) {def_pmax[0]=atof_dpi(ift.item[ai].value); n++;}
    /* k2 */
    strcpy(tmp, "k2_lower"); ai=iftGet(&ift, tmp);
    if(ai>=0) {def_pmin[1]=atof_dpi(ift.item[ai].value); n++;}
    strcpy(tmp, "k2_upper"); ai=iftGet(&ift, tmp);
    if(ai>=0) {def_pmax[1]=atof_dpi(ift.item[ai].value); n++;}
    /* BP */
    strcpy(tmp, "BP_lower"); ai=iftGet(&ift, tmp);
    if(ai>=0) {def_pmin[2]=atof_dpi(ift.item[ai].value); n++;}
    strcpy(tmp, "BP_upper"); ai=iftGet(&ift, tmp);
    if(ai>=0) {def_pmax[2]=atof_dpi(ift.item[ai].value); n++;}
    iftEmpty(&ift);
    if(TEST) printf("%d constraints read from %s\n", n, limfile);
    if(n<1) {
      fprintf(stderr, "Error: invalid parameter constraint file.\n");
      return(9);
    }
    /* Check that these are ok */
    for(pi=n=0, ret=0; pi<parNr; pi++) {
      if(def_pmin[pi]<0.0) ret++;
      if(def_pmax[pi]<def_pmin[pi]) ret++;
      if(def_pmax[pi]>def_pmin[pi]) n++;
    }
    if(ret) {
      fprintf(stderr, "Error: invalid parameter constraints.\n");
      return(9);
    }
    if(n==0) {
      fprintf(stderr, "Error: no model parameters left free for fitting.\n");
      return(9);
    }
  }
  if(TEST) {
    printf("Parameter constraints:\n");
    for(pi=0; pi<parNr; pi++) {
      printf("def_pmin[%d] := %g\n", pi+1, def_pmin[pi]);
      printf("def_pmax[%d] := %g\n", pi+1, def_pmax[pi]);
    }
  }


  /*
   *  Read data file
   */
  if(VERBOSE) printf("reading data file\n");
  dftInit(&dft);
  if(dftRead(dftfile, &dft)) {
    fprintf(stderr, "Error in reading '%s': %s\n", dftfile, dfterrmsg);
    return(2);
  }
  if(TEST>20) dftPrint(&dft);
  /* Check for NA's */
  if(dft_nr_of_NA(&dft) > 0) {
    fprintf(stderr, "Error: missing sample(s) in %s\n", dftfile);
    dftEmpty(&dft); return(2);
  }
  /* Set time unit to min */
  ret=dftTimeunitConversion(&dft, TUNIT_MIN);
  if(ret) fprintf(stderr, "Warning: check that regional data times are in minutes.\n");
  /* Remove frame overlaps and gaps */
  if(dft.timetype==3) {
    if(VERBOSE) fprintf(stdout, "checking frame overlap in %s\n", dftfile);
    ret=dftDeleteFrameOverlap(&dft);
    if(ret) {
      fprintf(stderr, "Error: %s has overlapping frame times.\n", dftfile);
      dftEmpty(&dft); return(2);
    }
  }
  /* Set fit duration */
  starttime=0; endtime=fitdur;
  fitframeNr=fittime_from_dft(&dft, &starttime, &endtime, &first, &last);
  if(fitframeNr<4) {
    fprintf(stderr, "Error: too few data points for a decent fit.\n");
    dftEmpty(&dft); return(2);
  }
  if(TEST) {
    printf("dft.frameNr := %d\n", dft.frameNr);
    printf("starttime := %g\n", starttime);
    printf("endtime := %g\n", endtime);
    printf("first := %d\n", first);
    printf("last := %d\n", last);
    printf("fitframeNr := %d\n", fitframeNr);
  }
  fitdur=endtime;
  /* Check that there is not any significant delay in the beginning of the data */
  if(dft.timetype==3) {
    if(dft.x1[0]>0.45) {
      fprintf(stderr, "Error: TACs must start at time zero.\n");
      dftEmpty(&dft); return(2);
    }
    if(dft.x1[0]>0.0833333) {
      fprintf(stderr, "Warning: TACs should start at time zero.\n");
    }
  }
  if(TEST>1) printf("Tissue calibration unit := %s\n", dft.unit);
  /* Check weighting */
  if(dft.isweight==0 && SILENT==0)
    fprintf(stderr, "Warning: data is not weighted.\n");
  /* Print the weights */
  if(TEST) {
    fprintf(stdout, "common_data_weights := %g", dft.w[0]);
    for(fi=1; fi<dft.frameNr; fi++) fprintf(stdout, ", %g", dft.w[fi]);
    fprintf(stdout, "\n");
  }

  /*
   *  Read reference TAC
   */
  if(VERBOSE) printf("reading reference\n");
  ret=dftReadReference(&dft, reffile, &inputtype, &ref, tmp);
  if(ret<=0) {
    fprintf(stderr, "Error in reading reference input: %s\n", tmp);
    dftEmpty(&dft); if(TEST) printf("ret := %d\n", ret);
    return(3);
  }
  if(ret>1) {
    fprintf(stderr, "Warning: several reference regions found: %s selected.\n",
            dft.voi[ref].name);
  } else if(VERBOSE) printf("reference_region := %s\n", dft.voi[ref].name);
  if(TEST) printf("inputtype := %d\n", inputtype);


  /* Calculate tissue integrals to determine constraints for BP value */
  if(!limfile[0]) {
    if(dft.timetype==3) for(ri=0; ri<dft.voiNr; ri++)
      petintegrate(dft.x1, dft.x2, dft.voi[ri].y, fitframeNr, dft.voi[ri].y3, NULL);
    else for(ri=0; ri<dft.voiNr; ri++)
      integrate(dft.x, dft.voi[ri].y, fitframeNr, dft.voi[ri].y3);
  }


  /*
   *  Prepare the room for results
   */
  if(VERBOSE) printf("preparing results\n");
  resInit(&res); ret=res_allocate_with_dft(&res, &dft); if(ret!=0) {
    fprintf(stderr, "Error: cannot setup memory for results.\n");
    dftEmpty(&dft); return(6);
  }
  /* Copy titles & filenames */
  strcpy(res.program, PROGRAM);
  strcpy(res.datafile, dftfile);
  if(inputtype!=5 && reffile[0]) strcpy(res.reffile, reffile);
  strcpy(res.refroi, dft.voi[ref].name);
  strcpy(res.fitmethod, "TGO");
  /* Set data range */
  sprintf(res.datarange, "%g - %g %s", 0.0, fitdur, petTunit(dft.timeunit));
  res.datanr=fitframeNr;
  /* Set current time to results */
  res.time=time(NULL);
  res.isweight=dft.isweight;
  /* Set parameter names */
  res.parNr=5;
  if(doDVR==0) strcpy(res.titleline, "R1 k2 BP WSS AIC");
  else strcpy(res.titleline, "R1 k2 DVR WSS AIC");
  strcpy(res.unitline, ". 1/min . . .");


  /*
   *  Fit one VOI at a time
   */
  if(VERBOSE) printf("fitting\n");
  /* Set common data pointers */
  t=dft.x;
  cr=dft.voi[ref].y;
  refIntegral=dft.voi[ref].y3[fitframeNr-1];
  /* Fit model to one TAC at a time */
  for(ri=0; ri<dft.voiNr; ri++) if(ri!=ref) {

    if(VERBOSE) printf("Region %d %s\n", ri+1, dft.voi[ri].name);
    /* Set data pointers */
    tis=dft.voi[ri].y; ct=dft.voi[ri].y2; w=dft.w;
    p=res.voi[ri].parameter;

    /* Set common parameter constraints */
    for(pi=0; pi<parNr; pi++) {pmin[pi]=def_pmin[pi]; pmax[pi]=def_pmax[pi];}
    /* Set BP limits based on the integrals of roi and ref TACs
       if those were not given in limfile */
    if(!limfile[0] && refIntegral>0.0) {
      a=(dft.voi[ri].y3[fitframeNr-1]/refIntegral);
      if(a<1.0) a=1.0; b=0.0*a; c=5.0*a; pmin[2]=b; pmax[2]=c;
    }
    if(TEST>3) {
      printf("Parameter constraints:\n");
      for(pi=0; pi<parNr; pi++) {
        printf("  %10.3E - %10.3E\n", pmin[pi], pmax[pi]);
      }
    }

    /* Fit */
    if(TEST>2) printf("  fitting curve...\n");
    for(pi=0; pi<parNr; pi++) p[pi]=0.0; p[parNr]=0.0;
    ret=tgo(
      pmin, pmax, srtm_func, parNr, 5, p+parNr, p, 0);
    if(ret>0) {
      fprintf(stderr, "\nError in optimization (%d).\n", ret);
      dftEmpty(&dft); resEmpty(&res); return(8);
    }
    if(TEST>3) {
      for(pi=0; pi<parNr; pi++) printf(" %g", p[pi]);
      printf(" -> WSS=%g\n", p[parNr]); fflush(stdout);
    }
    /* Correct fitted parameters to match constraints like inside the function */
    (void)modelCheckParameters(parNr, pmin, pmax, p, p, NULL);

    /* Calculate AIC, based on nr of parameters that actually are fitted */
    if(TEST>1) printf("calculating AIC etc.\n");
    for(pi=n=0; pi<parNr; pi++) if(pmax[pi]>pmin[pi]) n++;
    wss=wss_wo_penalty; p[res.parNr-1]=aic=aicSS(wss, fitframeNr, n);
    if(TEST>2) printf("wss := %g\nn := %d\nfitframeNr := %d\naic := %g\n",
      wss, n, fitframeNr, aic);

    /* Boostrap */
    if(doBootstrap) {
      tis=dft.voi[ref].y2; ct=dft.voi[ref].y3;
      for(fi=0; fi<fitframeNr; fi++) tis[fi]=dft.voi[ri].y[fi];
      if(doSD) sd=res.voi[ri].sd; else sd=NULL;
      if(doCL) {cl1=res.voi[ri].cl1; cl2=res.voi[ri].cl2;} else cl1=cl2=NULL;
      ret=bootstrap(
        cl1, cl2, sd,
        p, pmin, pmax, fitframeNr, tis, dft.voi[ri].y2, parNr, w, srtm_func
      );
      if(ret) {
        fprintf(stderr, "Error in bootstrap: %s\n", bserrmsg);
        for(pi=0; pi<parNr; pi++) {
          if(doSD) sd[pi]=nan(""); if(doCL) cl1[pi]=cl2[pi]=nan("");
        }
      }
    }

    /* Set very small values to zero */
    for(pi=0; pi<parNr; pi++) if(fabs(p[pi])<1.0E-8) p[pi]=0.0;

  } /* Next VOI */
  if(SILENT==0) {fprintf(stdout, "\n"); fflush(stdout);}

  /* Delete reference region(s) from the results and from the data,
     unless it already existed in data */
  if(inputtype==5) {
    resDelete(&res, ref);
  } else {
    for(ri=dft.voiNr-1; ri>=0; ri--) if(dft.voi[ri].sw!=0) {
      resDelete(&res, ri); dftDelete(&dft, ri);
    }
    ref=-1;
  }

  /*
   *  Convert BP to DVR, if necessary
   */
  if(doDVR!=0) {
    if(VERBOSE) printf("converting BP to DVR\n");
    for(ri=0; ri<res.voiNr; ri++) {
      res.voi[ri].parameter[2]+=1.0;
      if(doCL && !isnan(res.voi[ri].cl1[2])) res.voi[ri].cl1[2]+=1.0;
      if(doCL && !isnan(res.voi[ri].cl2[2])) res.voi[ri].cl2[2]+=1.0;
    }
  }

  /*
   *  Print results on screen
   */
  resPrint(&res);

  /*
   *  Save results
   */
  if(VERBOSE) printf("saving results in %s\n", resfile);
  ret=resWrite(&res, resfile);
  if(ret) {
    fprintf(stderr, "Error in writing '%s': %s\n", resfile, reserrmsg);
    resEmpty(&res); dftEmpty(&dft);
    return(11);
  }


  /*
   *  Saving and/or plotting of fitted TACs
   */
  if(svgfile[0] || fitfile[0]) {

    /* Create a DFT containing fitted TACs */
    char tmp[64];
    DFT dft2;
    dftInit(&dft2); ret=dftdup(&dft, &dft2);
    if(ret) {
      fprintf(stderr, "Error: cannot save fitted curves.\n");
      dftEmpty(&dft); resEmpty(&res);
      return(21);
    }
    for(ri=0; ri<dft.voiNr; ri++) if(ri!=ref) for(fi=0; fi<fitframeNr; fi++)
      dft2.voi[ri].y[fi]=dft2.voi[ri].y2[fi];
    dft2.frameNr=fitframeNr;

    /* Save fitted TACs */
    if(fitfile[0]) {
      if(VERBOSE) printf("saving fitted curves\n");
      sprintf(dft2.comments, "# program := %s\n", PROGRAM);    
      if(dftWrite(&dft2, fitfile)) {
        fprintf(stderr, "Error in writing '%s': %s\n", fitfile, dfterrmsg);
        dftEmpty(&dft2); dftEmpty(&dft); resEmpty(&res);
        return(22);
      }
      if(SILENT==0) printf("Fitted TACs written in %s\n", fitfile);
    }

    /* Save SVG plot of fitted and original data */
    if(svgfile[0]) {
      if(VERBOSE) printf("saving SVG plot\n");
      sprintf(tmp, "SRTM fit ");
      if(strlen(dft.studynr)>1) strcat(tmp, dft.studynr);
      ret=plot_fit_svg(&dft, &dft2, tmp, svgfile);
      if(ret) {
        fprintf(stderr, "Error (%d) in writing '%s'.\n", ret, svgfile);
        dftEmpty(&dft2); dftEmpty(&dft); resEmpty(&res);
        return(30+ret);
      }
      if(SILENT==0) printf("Plots written in %s\n", svgfile);
    }

    dftEmpty(&dft2);
  }

  /* Free memory */
  dftEmpty(&dft); resEmpty(&res);

  return(0);
}
/*****************************************************************************/

/******************************************************************************
 *
 *  Function to be minimized
 *
 ******************************************************************************/
double srtm_func(double *p)
{
  int i, ret=0;
  double R1, k2, BP, wss, d;
  double pa[MAX_PARAMETERS], penalty=1.0;

  /* Check parameters against the constraints */
  ret=modelCheckParameters(parNr, pmin, pmax, p, pa, &penalty);
  /* Get parameters */
  wss=0.0; R1=pa[0]; k2=pa[1]; BP=pa[2];

  /* Simulate tissue curves */
  ret=simSRTM(t, cr, fitframeNr, R1, k2, BP, ct);
  if(ret) {
    if(TEST>2) printf("  error %d in simulation\n", ret);
    return(BAD_FIT);
  }

  /* Calculate error */
  for(i=0, wss=0; i<fitframeNr; i++) if(w[i]>0.0) {
    d = ct[i] - tis[i];
    wss += w[i]*d*d;
  }
  wss_wo_penalty=wss;
  wss*=penalty;
  if(TEST>10) printf("%g  %g  %g => %g\n", R1, k2, BP, wss);
  return(wss);
}
/******************************************************************************/

/******************************************************************************
 *
 *  void print_usage()
 *
 ******************************************************************************/
void print_usage()
{
  int i=0;
  static char *info[] = {
  " ",
  PROGRAM,
  " ",
  "Estimates R1 (=K1/K1'), k2 and BPnd (binding potential) using simplified",
  "reference tissue compartment model, SRTM (1,2,3). Assumptions are that",
  "K1/k2 is the same in all brain regions, and that a one-tissue compartment",
  "model with plasma input could fit the tissue curves satisfactorily.",
  "This program uses nonlinear regression in model fitting.",
  " ",
  "Parameters:",
  "  1) Tissue TAC file (4)",
  "  2) Name of reference region in tissue file, or name of file",
  "     containing reference tissue TAC",
  "  3) Fit time (in minutes from injection)",
  "  4) Result file (existing file is overwritten) (5)",
  " ",
  "Options:",
  " -DVR",
  "     Instead of BPnd, program saves the DVR (=BPnd+1) values.",
  " -lim=<filename>",
  "     Specify the constraints for model parameters;",
  "     This file with default values can be created by giving this",
  "     option as the only command-line argument to this program.",
  " -SD[=<y|N>]",
  "     Standard deviations are calculated and saved in results (y), or",
  "     not calculated (n).",
  " -CL[=<y|N>]",
  "     95% Confidence limits are calculated and saved in results (y), or",
  "     not calculated (n).",
  " -fit=<Filename>",
  "     Fitted regional TACs are written in DFT format (4).",
  " -svg=<Filename>",
  "     Fitted and measured TACs are plotted in specified SVG file (6).",
  " -h or --help",
  "     Print this message and exit.",
  " --version or --build",
  "     Print software build information and exit.",
  " --silent",
  "     Program works silently, printing only error and warning messages.",
  " --verbose",
  "     Program prints more information about what it is doing.",
  " ",
  "If datafile contains weights, those are used in the fitting.",
  "Values of R1, k2, and BPnd are written in the specified result file.",
  "Fitted curves are written in DFT format, if filename is given.",
  " ",
  "Example 1: file a789.dft contains regions-of-interest and reference region,",
  "with name 'cer all'. The whole time range is used in the fit.",
  "     fit_srtm a789.dft 'cer all' 999 a789.res",
  " ",
  "Example 2: Reference region TAC is in a separate file, a789ref.dft.",
  "Standard deviations and confidence limits are also estimated.",
  "     fit_srtm -SD=y -CL=y a789.dft a789ref.dft 999 a789.res",
  " ",
  "References:",
  "1. Cunningham VJ, Hume SP, Price GR, Ahier RG, Cremer JE, Jones AKP.",
  "   Compartmental analysis of diprenorphine binding to opiate receptors",
  "   in the rat in vivo and its comparison with equilibrium data in vitro.",
  "   J Cereb Blood Flow Metab 1991;11:1-9.",
  "2. Lammertsma AA, Hume SP. Simplified reference tissue model for PET",
  "   receptor studies. Neuroimage 1996;4:153-158.",
  "3. Oikonen V, Sederholm K. TPCMOD0002: Model equations for reference tissue",
  "   compartmental models. http://www.turkupetcentre.net/reports/tpcmod0002.pdf",
  "4. File format specification: DFT.",
  "   http://www.turkupetcentre.net/formats/format_dft_1_0_0.pdf",
  "5. File format specification: Result file.",
  "   http://www.turkupetcentre.net/formats/format_res_1_0_0.pdf",
  "6. Scalable Vector Graphics (SVG) 1.1 Specification.",
  "   http://www.w3.org/TR/SVG/",
  " ",
  "See also: dftweigh, rescoll, logan, regbfbp, fit_frtm, r2t_rtcm",
  " ",
  "Keywords: DFT, modelling, binding potential, BP, SRTM, reference input",
  " ",
  "This program comes with ABSOLUTELY NO WARRANTY. This is free software, and",
  "you are welcome to redistribute it under GNU General Public License.",
  " ",
  0};
  while(info[i]!=0) printf("  %s\n", info[i++]);
}
/******************************************************************************/

/******************************************************************************/
/**
 * Prints build information to stdout.
 */
void print_build() {

  printf("\n %s\n", PROGRAM);
  printf("\n Build %s %s\n\n",__DATE__,__TIME__);
#if defined(__FILE__) && defined(__TIMESTAMP__)
  printf(" File %s last edited on %s\n", __FILE__, __TIMESTAMP__);
#endif
  printf(" _______________________________________________\n");

  libtpcmodel_print_build(stdout);
  libtpcmodext_print_build(stdout);
  libtpccurveio_print_build(stdout);
  libtpcmisc_print_build(stdout);
  libtpcsvg_print_build(stdout);
  printf(" _______________________________________________\n\n");

#if defined(__STDC_VERSION__)
  printf(" Version of C: %ld\n", __STDC_VERSION__);
  printf(" _______________________________________________\n\n");
#endif
#if defined(__STDC_ISO_10646__)
  printf(" Compiler ISO/IEC 10646 specification: %ld\n", __STDC_ISO_10646__);
  printf(" _______________________________________________\n\n");
#endif
#if defined(__GNUC__) && defined(__VERSION__)
  printf(" GNU C version: %s\n", __VERSION__);
  printf(" _______________________________________________\n\n");
#endif
}
/*****************************************************************************/

/*****************************************************************************/
